;
; md5sum.asm
;
; Written by
;  Kajtar Zsolt <soci@c64.rulez.org>
;
;; md5sum [files]
;;
;;   Calculates md5 hash on files as described in RFC 1321, and prints
;;   the result. It can be used for file integrity checking, better then
;;   cksum, but much slower.
;;
;    This file is part of the IDE64 shell utilities.
;    See README for copyright notice.
;
;    This program is free software; you can redistribute it and/or modify
;    it under the terms of the GNU General Public License as published by
;    the Free Software Foundation; either version 2 of the License, or
;    (at your option) any later version.
;
;    This program is distributed in the hope that it will be useful,
;    but WITHOUT ANY WARRANTY; without even the implied warranty of
;    MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
;    GNU General Public License for more details.
;
;    You should have received a copy of the GNU General Public License
;    along with this program; if not, write to the Free Software
;    Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA  02111-1307  USA
;

WILDEXP		= 1

		.logical 2
aa		.fill 4
bb		.fill 4
cc		.fill 4
dd		.fill 4
		.here

bufp		= $22
ichars		= $24
xx		= $25
len		= $ac

banner		= "md5sum 0.1 by soci/singular"

		.include "lib.asm"

		.section code
start2		ldx #<arglist
		ldy #>arglist
		jsr parseargs
		lda #<arglist
		sta p1
		lda #>arglist
		sta p1+1

loop		ldy #0
		lda (p1),y
		bne +
q       	jmp exit

+		ldx p1
		ldy p1+1
		pha
                jsr addstring
		pla
		inx
		bne +
		iny
+		jsr onefile
		ldx $91
		cpx #$7f
		beq q
		jmp loop

onefile		jsr $ffbd
		lda #1
		ldx mydrive
		ldy #0
		jsr $ffba
		jsr $ffc0

		ldx #15
-		lda aa,x
		pha
		lda initval,x
		sta aa,x
		cpx #4
		bge +
		lda #0
		sta xx
		sta len,x
+		dex
		bpl -

		jsr in

		ldx #240
-		lda <(aa-240),x
		sta result-240,x
		pla
		sta <(aa-240),x
		inx
		bne -
		bcc +
                rts
+
		ldx #3
		jsr $ffc9

		ldx #240
-		lda result-240,x
		jsr printhex
		inx
		bne -
		lda #32
		jsr $ffd2
		jsr $ffd2
		ldy #0
-		cpy $b7
		beq +
		lda ($bb),y
		jsr $ffd2
		iny
		bne -
+		lda #13
		jsr $ffd2
		jmp $ffcc

in		.proc
		ldx #1
		stx myin
		jsr $ffc6

next            clc
		lda $90
                bne ujj
		lda #<buffer
		sta bufp
		lda #>buffer
		sta bufp+1

		ldx #<size(buffer)
		ldy #>size(buffer)
		lda #bufp
		jsr read
		bcs ujj
		txa
                adc len
                sta len
                tya
                adc len+1
                sta len+1
                bcc +
		inc len+2
                bne +
                inc len+3
+
		stx ichars
		tya
		asl ichars
		rol
		asl ichars
		rol
		beq ur
-		pha
		jsr md5
		lda bufp
		clc
		adc #64
		sta bufp
		bcc +
		inc bufp+1
+		pla
		sec
		sbc #1
		bne -
ur
		lda $90
		beq next

		lda ichars
		lsr
		lsr
		pha
		tay
		lda #$80
-		sta (bufp),y
		lda #0
		iny
		cpy #64
		blt -
		pla
		cmp #56
		blt +
		jsr md5
		ldy #63
		lda #0
-		sta (bufp),y
		dey
		bpl -

+		ldy #56
-		lda len-56,y
		sta (bufp),y
		iny
		cpy #60
		blt -
		ldx #3
-		ldy #56
		clc
-		lda (bufp),y
		rol
		sta (bufp),y
		iny
		tya
		eor #61
		bne -
		dex
		bne --
		jsr md5
		clc

ujj             php
		jsr $ffcc
		lda #1
		jsr $ffc3
                plp
                rts
		.pend

shift		.macro
		.if \2=9
		tay
		asl
		lda \1
		rol
		ldx \1+1
		sta \1+1
		txa
		rol
		ldx \1+2
		sta \1+2
		txa
		rol
		sta \1+3
		tya
		rol
		sta \1
		.elsif \2=23
		tay
		lsr
		lda \1+2
		ror
		ldx \1+1
		sta \1+1
		txa
		ror
		ldx \1
		sta \1
		txa
		ror
		sta \1+3
		tya
		ror
		sta \1+2
		.elsif \2>16
		ldy #32-\2
-		lsr
		ror \1+2
		ror \1+1
		ror \1
		bcc +
		ora #128
+		dey
		bne -
		sta \1+3
		.else
		ldy #\2
-		asl
		rol \1
		rol \1+1
		rol \1+2
		adc #0
		dey
		bne -
		sta \1+3
		.fi
		.endm

stepf		.macro
		clc
		lda \4+0
		eor \3+0
		and \2+0
		eor \4+0
		adc \1+0
		rol xx
		ldy #\5*4+0
		adc (bufp),y
		rol xx
		adc #((\6) >> (0*8)) & $ff
		sta \1+0

		lda \4+1
		eor \3+1
		and \2+1
		eor \4+1
		adc \1+1
		ror xx
		iny
		adc (bufp),y
		ror xx
		adc #((\6) >> (1*8)) & $ff
		sta \1+1

		lda \4+2
		eor \3+2
		and \2+2
		eor \4+2
		adc \1+2
		rol xx
		iny
		adc (bufp),y
		rol xx
		adc #((\6) >> (2*8)) & $ff
		sta \1+2

		lda \4+3
		eor \3+3
		and \2+3
		eor \4+3
		adc \1+3
		lsr xx
		iny
		adc (bufp),y
		lsr xx
		adc #((\6) >> (3*8)) & $ff

		#shift \1, \7
		clc
		.for i=0,i<4,i=i+1
		lda \1+i
		adc \2+i
		sta \1+i
		.next
		.endm

stepg		.macro
		clc
		lda \3+0
		eor \2+0
		and \4+0
		eor \3+0
		adc \1+0
		rol xx
		ldy #\5*4+0
		adc (bufp),y
		rol xx
		adc #((\6) >> (0*8)) & $ff
		sta \1+0

		lda \3+1
		eor \2+1
		and \4+1
		eor \3+1
		adc \1+1
		ror xx
		iny
		adc (bufp),y
		ror xx
		adc #((\6) >> (1*8)) & $ff
		sta \1+1

		lda \3+2
		eor \2+2
		and \4+2
		eor \3+2
		adc \1+2
		rol xx
		iny
		adc (bufp),y
		rol xx
		adc #((\6) >> (2*8)) & $ff
		sta \1+2

		lda \3+3
		eor \2+3
		and \4+3
		eor \3+3
		adc \1+3
		lsr xx
		iny
		adc (bufp),y
		lsr xx
		adc #((\6) >> (3*8)) & $ff

		#shift \1, \7
		clc
		.for i=0,i<4,i=i+1
		lda \1+i
		adc \2+i
		sta \1+i
		.next
		.endm

steph		.macro
		clc
		lda \2+0
		eor \3+0
		eor \4+0
		adc \1+0
		rol xx
		ldy #\5*4+0
		adc (bufp),y
		rol xx
		adc #((\6) >> (0*8)) & $ff
		sta \1+0

		lda \2+1
		eor \3+1
		eor \4+1
		adc \1+1
		ror xx
		iny
		adc (bufp),y
		ror xx
		adc #((\6) >> (1*8)) & $ff
		sta \1+1

		lda \2+2
		eor \3+2
		eor \4+2
		adc \1+2
		rol xx
		iny
		adc (bufp),y
		rol xx
		adc #((\6) >> (2*8)) & $ff
		sta \1+2

		lda \2+3
		eor \3+3
		eor \4+3
		adc \1+3
		lsr xx
		iny
		adc (bufp),y
		lsr xx
		adc #((\6) >> (3*8)) & $ff

		#shift \1, \7
		clc
		.for i=0,i<4,i=i+1
		lda \1+i
		adc \2+i
		sta \1+i
		.next
		.endm

stepi		.macro
		clc
		lda \4+0
		eor #255
		ora \2+0
		eor \3+0
		adc \1+0
		rol xx
		ldy #\5*4+0
		adc (bufp),y
		rol xx
		adc #((\6) >> (0*8)) & $ff
		sta \1+0

		lda \4+1
		eor #255
		ora \2+1
		eor \3+1
		adc \1+1
		ror xx
		iny
		adc (bufp),y
		ror xx
		adc #((\6) >> (1*8)) & $ff
		sta \1+1

		lda \4+2
		eor #255
		ora \2+2
		eor \3+2
		adc \1+2
		rol xx
		iny
		adc (bufp),y
		rol xx
		adc #((\6) >> (2*8)) & $ff
		sta \1+2

		lda \4+3
		eor #255
		ora \2+3
		eor \3+3
		adc \1+3
		lsr xx
		iny
		adc (bufp),y
		lsr xx
		adc #((\6) >> (3*8)) & $ff

		#shift \1, \7
		clc
		.for i=0,i<4,i=i+1
		lda \1+i
		adc \2+i
		sta \1+i
		.next
		.endm

md5		.proc

		.for i=15,i>=0,i=i-1
		lda aa+i
		pha
		.next

		#stepf aa, bb, cc, dd, 0, -680876936, 7
		#stepf dd, aa, bb, cc, 1, -389564586, 12
		#stepf cc, dd, aa, bb, 2, $242070db, 17
		#stepf bb, cc, dd, aa, 3, -1044525330, 22
		#stepf aa, bb, cc, dd, 4, -176418897, 7
		#stepf dd, aa, bb, cc, 5, $4787c62a, 12
		#stepf cc, dd, aa, bb, 6, -1473231341, 17
		#stepf bb, cc, dd, aa, 7, -45705983, 22
		#stepf aa, bb, cc, dd, 8, $698098d8, 7
		#stepf dd, aa, bb, cc, 9, -1958414417, 12
		#stepf cc, dd, aa, bb, 10, -42063, 17
		#stepf bb, cc, dd, aa, 11, -1990404162, 22
		#stepf aa, bb, cc, dd, 12, $6b901122, 7
		#stepf dd, aa, bb, cc, 13, -40341101, 12
		#stepf cc, dd, aa, bb, 14, -1502002290, 17
		#stepf bb, cc, dd, aa, 15, $49b40821, 22

		#stepg aa, bb, cc, dd, 1, -165796510, 5
		#stepg dd, aa, bb, cc, 6, -1069501632, 9
		#stepg cc, dd, aa, bb, 11, $265e5a51, 14
		#stepg bb, cc, dd, aa, 0, -373897302, 20
		#stepg aa, bb, cc, dd, 5, -701558691, 5
		#stepg dd, aa, bb, cc, 10, $02441453, 9
		#stepg cc, dd, aa, bb, 15, -660478335, 14
		#stepg bb, cc, dd, aa, 4, -405537848, 20
		#stepg aa, bb, cc, dd, 9, $21e1cde6, 5
		#stepg dd, aa, bb, cc, 14, -1019803690, 9
		#stepg cc, dd, aa, bb, 3, -187363961, 14
		#stepg bb, cc, dd, aa, 8, $455a14ed, 20
		#stepg aa, bb, cc, dd, 13, -1444681467, 5
		#stepg dd, aa, bb, cc, 2, -51403784, 9
		#stepg cc, dd, aa, bb, 7, $676f02d9, 14
		#stepg bb, cc, dd, aa, 12, -1926607734, 20

		#steph aa, bb, cc, dd, 5, -378558, 4
		#steph dd, aa, bb, cc, 8, -2022574463, 11
		#steph cc, dd, aa, bb, 11, $6d9d6122, 16
		#steph bb, cc, dd, aa, 14, -35309556, 23
		#steph aa, bb, cc, dd, 1, -1530992060, 4
		#steph dd, aa, bb, cc, 4, $4bdecfa9, 11
		#steph cc, dd, aa, bb, 7, -155497632, 16
		#steph bb, cc, dd, aa, 10, -1094730640, 23
		#steph aa, bb, cc, dd, 13, $289b7ec6, 4
		#steph dd, aa, bb, cc, 0, -358537222, 11
		#steph cc, dd, aa, bb, 3, -722521979, 16
		#steph bb, cc, dd, aa, 6, $04881d05, 23
		#steph aa, bb, cc, dd, 9, -640364487, 4
		#steph dd, aa, bb, cc, 12, -421815835, 11
		#steph cc, dd, aa, bb, 15, $1fa27cf8, 16
		#steph bb, cc, dd, aa, 2, -995338651, 23

		#stepi aa, bb, cc, dd, 0, -198630844, 6
		#stepi dd, aa, bb, cc, 7, $432aff97, 10
		#stepi cc, dd, aa, bb, 14, -1416354905, 15
		#stepi bb, cc, dd, aa, 5, -57434055, 21
		#stepi aa, bb, cc, dd, 12, $655b59c3, 6
		#stepi dd, aa, bb, cc, 3, -1894986606, 10
		#stepi cc, dd, aa, bb, 10, -1051523, 15
		#stepi bb, cc, dd, aa, 1, -2054922799, 21
		#stepi aa, bb, cc, dd, 8, $6fa87e4f, 6
		#stepi dd, aa, bb, cc, 15, -30611744, 10
		#stepi cc, dd, aa, bb, 6, -1560198380, 15
		#stepi bb, cc, dd, aa, 13, $4e0811a1, 21
		#stepi aa, bb, cc, dd, 4, -145523070, 6
		#stepi dd, aa, bb, cc, 11, -1120210379, 10
		#stepi cc, dd, aa, bb, 2, $2ad7d2bb, 15
		#stepi bb, cc, dd, aa, 9, -343485551, 21

		clc
		.for i=0,i<4,i=i+1
		pla
		adc aa+i
		sta aa+i
		.next
		clc
		.for i=0,i<4,i=i+1
		pla
		adc bb+i
		sta bb+i
		.next
		clc
		.for i=0,i<4,i=i+1
		pla
		adc cc+i
		sta cc+i
		.next
		clc
		.for i=0,i<4,i=i+1
		pla
		adc dd+i
		sta dd+i
		.next
		rts
		.pend
		.send code

initval		.section data
		.byte $01,$23,$45,$67
		.byte $89,$ab,$cd,$ef
		.byte $fe,$dc,$ba,$98
		.byte $76,$54,$32,$10
		.send data

		.section bss
result		.fill 16

		.align 256
buffer		.fill $3000
		.send bss

