;
; sh.asm
;
; Written by
;  Kajtar Zsolt <soci@c64.rulez.org>
;
; Shell main file 0.2
;
;    This file is part of the IDE64 shell utilities.
;    See README for copyright notice.
;
;    This program is free software; you can redistribute it and/or modify
;    it under the terms of the GNU General Public License as published by
;    the Free Software Foundation; either version 2 of the License, or
;    (at your option) any later version.
;
;    This program is distributed in the hope that it will be useful,
;    but WITHOUT ANY WARRANTY; without even the implied warranty of
;    MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
;    GNU General Public License for more details.
;
;    You should have received a copy of the GNU General Public License
;    along with this program; if not, write to the Free Software
;    Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA  02111-1307  USA
;
;
;! IDEDOS shell
;!
;! This package provides some handy utilities to be used by
;! the shell feature of IDEDOS.
;!
;! The shell is invoked by the "#" character. This loads the shell executor
;! from the file "1//:sh,p" on the system drive.
;!
;! The first parameter after the "#" character
;! is the name of the file which will be loaded from from the system drive
;! from the directory "1//bin/:". Parameters are separated by space, unless
;! put in quotes.
;!
;! Shell utilities have special signature and are executed at an alternate
;! entry point to allow parsing the rest of the parameters. Normal programs
;! are executed by simply running them, thereby allowing a quickstart
;! possibility for frequently used utilities from any location.
;!
;! The shell utilities accept output redirection in form of ">filename",
;! which will write the output to the named file. The ">>filename" form will
;! append to an existing file.
;!
;! Some commands accept wildcards in the filenames, these will be expanded
;! to a list of files. This will only work for wildcards at the end of the path.
;!
;! The list of included utilities follows:
;!
		*=$334

                sec
		ldx #nameend-name
		ldy #0
                geq in

-		sta name,x
                inx
		eor #34                 ;quoted?
		bne +
                adc #1
                lsr
tr              dex
+               txa
                eor #nameend-name+17    ;too long name?
                beq tr
		inc $7a
		bne in
		inc $7b
in              lda ($7a),y
                beq kesz
                bcc -                   ;in quote mode...
		cmp #32
                sec
		bne -

kesz		lda #$80
		cpx #nameend-name
                beq null                ;no name...
                txa
		ldx #<name
		ldy #>name
		jsr $ffbd
		ldx $ba
		ldy #1
		jsr $ffba
		lda #0
		jsr $ffd5
null		pha
		lda $ff
		sta $ba
		pla
		bcs not
                tya
                pha
		dec $af
		ldy #246
-		lda ident-246,y
		cmp ($ae),y
		bne nm
		iny
		bne -
                pla
		lda #243-1
		adc $ae
		sta $ae
		bcc +
		inc $af
+		jsr do
nl		lda #1
		sta 198
		lda #"#"
		sta 631
		jmp ($302)

do		jmp ($ae)

nm              pla             ;normal program, run it
		inc $af
		stx $2d
		sta $2e
		ldy #5
		sty 198
-		lda run-1,y
		dey
		sta 631,y
		bne -
		jmp $a52a

not             tax
		bmi nl
		lda #1
		sta 198
		lda #"#"
		sta 631
		jmp ($300)	 ;exit with ready.

run		.text "run:{cr}"
ident		.text "ide64shell"

name		.text "1//bin/:"
nameend
