;
; uncpio.asm
;
; Written by
;  Kajtar Zsolt <soci@c64.rulez.org>
;
;; uncpio [files]
;;
;;   Extracts multiple cpio files, and creates the directory structure as
;;   well.
;;
;;   Can be used to install "packages" when used in the root directory as:
;;   #uncpio install/*
;;
;    This file is part of the IDE64 shell utilities.
;    See README for copyright notice.
;
;    This program is free software; you can redistribute it and/or modify
;    it under the terms of the GNU General Public License as published by
;    the Free Software Foundation; either version 2 of the License, or
;    (at your option) any later version.
;
;    This program is distributed in the hope that it will be useful,
;    but WITHOUT ANY WARRANTY; without even the implied warranty of
;    MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
;    GNU General Public License for more details.
;
;    You should have received a copy of the GNU General Public License
;    along with this program; if not, write to the Free Software
;    Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA  02111-1307  USA
;

WILDEXP		= 1

p2		= $22
p3		= $fd

banner		= "uncpio 0.2 by soci/singular"

		.include "lib.asm"

header_s	.struct
dev		.fill 2
ino		.fill 2
mode		.fill 2
uid		.fill 2
gid		.fill 2
nlink		.fill 2
rdev		.fill 2
mtime		.fill 4
namesize	.fill 2
filesize	.fill 4
name		.fill 256
		.ends

header2_s	.struct
dev		.fill 6
ino		.fill 6
mode		.fill 6
uid		.fill 6
gid		.fill 6
nlink		.fill 6
rdev		.fill 6
mtime		.fill 11
namesize	.fill 6
filesize	.fill 11
name		.fill 256
		.ends

header3_s	.struct
ino		.fill 8
mode		.fill 8
uid		.fill 8
gid		.fill 8
nlink		.fill 8
mtime		.fill 8
filesize	.fill 8
devmajor	.fill 8
devminor	.fill 8
rdevmajor	.fill 8
rdevminor	.fill 8
namesize	.fill 8
check		.fill 8
name		.fill 256
		.ends

		.section code
start2		ldx #<arglist
		ldy #>arglist
		jsr parseargs
		lda #<arglist
		sta p1
		lda #>arglist
		sta p1+1

		lda #15
		ldx mydrive
		tay
		jsr $ffba
		lda #0
		jsr $ffbd
		jsr $ffc0
		dec $1
		bcs err

loop		ldy #0
		lda (p1),y
		bne +
q		lda #15
		jsr $ffc3
err		inc $1
		jmp exit

+		ldx p1
		ldy p1+1
		pha
                jsr addstring
		pla
		inx
		bne +
		iny
+		jsr $ffbd
		lda #1
		ldx mydrive
		ldy #0
		jsr $ffba
		jsr $ffc0
		bcs +
		jsr onefile
		lda #1
		jsr $ffc3
+		ldx $91
		cpx #$7f
		beq q
		jmp loop

onefile		lda #0
		ldx #3
-		sta currentpos,x
		dex
		bpl -
		sta read.unsupp+1
		sta write.unsupp+1

;----------------------------------------------
sellp		lda #<header
		sta p2
		lda #>header
		sta p2+1

		ldx #1
		stx myin
		jsr $ffc6

		jsr read2
		cpx #$c7
		bne *+4
		cmp #$71
		beq binary
		cpx $71
		bne *+4
		cmp #$c7
		beq binary2
		cpx #"0"
		bne *+4
		cmp #"7"
		bne unk2
		jsr read2
		cpx #"0"
		bne *+4
		cmp #"7"
		bne unk2
		jsr read2
		cpx #"0"
		bne *+4
		cmp #"7"
		beq oldascii
		cmp #"1"
		beq newascii
		cmp #"2"
		beq newascii

unk2		ldx #3
		jsr $ffc9

		lda #<unknowntxt
		ldy #>unknowntxt
		inc $1
		jsr $ab1e
		dec $1
		jmp $ffcc

unknowntxt	.section data
		.null 13,"unknown format.",13
		.send data

newascii	lda #"0"
		ldx #7
-		sta header3.ino,x
		dex
		bpl -
		ldx #104
		ldy #0
		lda #p2
		jsr readx
		ldy #header3.ino-header3
		jsr getnum2
		stx header.ino
		tya
		ora nuum+2
		ora nuum+3
		sta header.ino+1
		ldy #header3.mode-header3
		jsr getnum2
		stx header.mode
		sty header.mode+1
		ldy #header3.namesize-header3
		jsr getnum2
		stx header.namesize
		sty header.namesize+1
		ldy #header3.filesize-header3
		jsr getnum2
		lda #3
		gne je

oldascii	lda #"0"
		ldx #5
-		sta header2.ino,x
		dex
		bpl -
		ldx #70
		ldy #0
		lda #p2
		jsr readx

		ldy #header2.ino-header2
		lda #header2.ino-header2+6
		jsr getnum
		stx header.ino
		sty header.ino+1
		ldy #header2.mode-header2
		lda #header2.mode-header2+6
		jsr getnum
		stx header.mode
		sty header.mode+1
		ldy #header2.namesize-header2
		lda #header2.namesize-header2+6
		jsr getnum
		stx header.namesize
		sty header.namesize+1
		ldy #header2.filesize-header2
		lda #header2.filesize-header2+11
		jsr getnum
		lda #0
je		stx header.filesize+2
		sty header.filesize+3
		tax
		lda nuum+2
		sta header.filesize+0
		lda nuum+3
		sta header.filesize+1
		jmp gp

binary2		clc
binary		ldx #24
		ldy #0
		sty header.ino
		sty header.ino+1
		lda #p2
		php
		jsr readx
		plp
		bcs +
		ldx #24
-		ldy header-2,x
		lda header-1,x
		sta header-2,x
		tya
		sta header-1,x
		dex
		dex
		bne -
+		ldx #1
gp		stx pad
		txa
		eor #$ff
		sta pad2

		lda #<header.name
		sta p2
		lda #>header.name
		sta p2+1
		ldx pad
		cpx #3
		bne +
		lda header.namesize	;newc
		adc #0
		and pad2
		tax
		lda header.namesize+1
		adc #0
		cpx #254
		inx
		inx
		jmp oj

+		lda header.namesize
		clc
		adc pad
		and pad2
		tax
		lda header.namesize+1
oj		adc #0
		tay
		lda #p2
		jsr readx

		lda header.ino
		ora header.ino+1
		bne +
		jmp $ffcc		; finished

+		lda #<buffer
		sta p2
		lda #>buffer
		sta p2+1

		jsr topetscii

		lda header.filesize
		sta nuum+2
		lda header.filesize+1
		sta nuum+3
		lda header.filesize+2
		sta nuum+0
		lda header.filesize+3
		sta nuum+1

		ldx #3
		jsr $ffc9
		bcs errstou

		ldy #0
-		lda #1
		sta $d4
		lda header.name,y
		beq +
		jsr $ffd2
		iny
		bne -
+		lsr $d4
		lda #"{cr}"
		jsr $ffd2
errstou
		jsr convertname

		lda header.mode+1
		and #$f0
		cmp #$40
		bne notdir
		ldx #15
		jsr $ffc9
		bcs errmd
		lda #"m"
		jsr $ffd2
		lda #"d"
		jsr $ffd2
		ldx #0
-		lda name,x
		beq +
		jsr $ffd2
		inx
		bne -
+		lda #13
		jsr $ffd2
		jsr $ffcc
errmd		jmp oe
notdir

normalfile
		lda #5
		ldx mydrive
		ldy #1
		jsr $ffba

		ldx #255
-		inx
		lda name,x
		bne -
		lda header.mode+1
		and #$f0
		cmp #$a0
		bne nlnk	      ;link?
		ldy #3
-		lda linktxt,y
		sta name,x
		inx
		dey
		bpl -

linktxt		.section data
		.text "knl,"
		.send data

		lda #0
		sta name,x
nlnk		txa
		ldx #<name
		ldy #>name
		jsr $ffbd
		jsr $ffc0

		lda header.mode+1
		and #$f0
		cmp #$a0
		bne nlnk2

		ldx #1
		jsr $ffc6		;Guess what, JiffyDos is buggy!

		lda #<header.name
		sta p2
		lda #>header.name
		sta p2+1
		lda nuum
		clc
		adc pad
		and pad2
		tax
		lda nuum+1
		adc #0
		tay
		lda #p2
		jsr readx
		ldx #5
		jsr $ffc9
		bcs errlnk

		jsr topetscii
		jsr convertname
		ldx #0
		lda name
		cmp #":"
		bne +
		inx
+
-		lda name,x
		beq +
		jsr $ffd2
		inx
		bne -
+		lda #0
		jsr $ffd2
		jsr $ffcc
errlnk		lda #5
		jsr $ffc3
		jmp oe
nlnk2

copylp		ldx #1
		jsr $ffc6		;Guess what, JiffyDos is buggy!
		bcs errcopy

		lda nuum
		clc
		adc pad
		and pad2
		tax
		lda nuum+1
		adc #0
		tay
		jsr calcsize
		beq endw
		lda #p2
		jsr readx
		stx v1+1
		sty v2+1
		tya
		pha

		ldx #5
		stx myout
		jsr $ffc9		;Guess what, JiffyDos is buggy!

		jsr calcsize2
		lda #p2
		jsr write
		jsr substract
		pla
		cmp #>size(buffer)
		beq copylp

endw		jsr $ffcc
errcopy		lda #5
		jsr $ffc3
oe		jmp sellp

read2		jsr $ffcf
		tax
		jmp $ffcf

substract	lda nuum
		sec
v1		sbc #0
		sta nuum
		lda nuum+1
v2		sbc #0
		sta nuum+1
		bcs +
		lda nuum+2
		bne kq
		dec nuum+3
kq		dec nuum+2
+		rts

calcsize2	.proc
		ldx nuum
		ldy nuum+1
		clc
		gcc calcsize
		.pend

calcsize	.proc
		bcs q
		lda nuum+3
		ora nuum+2
		bne q
		cpy #>size(buffer)
		blt +
q		ldx #0
		ldy #>size(buffer)
+		txa
		bne +
		tya
+		rts
		.pend
;-----------------------------------
readx		.proc
		jsr read
		txa
		clc
		adc currentpos
		sta currentpos
		tya
		adc currentpos+1
		sta currentpos+1
		bcc +
		inc currentpos+2
		bne +
		inc currentpos+3
+		rts
		.pend
;-----------------------------------
convertname	.proc
		ldy #0
-		lda header.name,y
		beq +
		cmp #":"
		beq kesz
		iny
		cpy #200
		blt -
+
		lda #"/"
-		dey
		cmp header.name,y
		beq -			;skip trailing "/"
		iny
		lda #0
		sta header.name,y

		lda #"/"
-		dey
		beq ve
		cmp header.name,y
		bne -			;any "/"?
		sta name
		lda #":"
		sta name+2,y
		tya
		tax
-		lda header.name,y
		sta name+1,y
		dey
		cpy #255
		bne -
-		inx
		cpx #200
		beq +
		lda header.name,x
		sta name+2,x
		bne -
+		lda #0
		sta name+2,x
		geq kesz

ve		lda header.name,y	;normal filename
		sta name+1,y
		beq +
		iny
		cpy #200
		blt ve
+		lda #0
		sta name+1,y
		lda #":"
		sta name
kesz		rts
		.pend
;-----------------------------------
getnum		.proc
		sta max+1
		ldx #3
		lda #0
-		sta nuum,x
		dex
		bpl -
lp		lda header,y
		eor #$30
		cmp #8
		bcs +
		lda #%00111100
-		ldx #252
-		rol nuum-252,x
		inx
		bne -
		lsr
		bcc --
		and header,y
		ora nuum
		sta nuum
+		iny
max		cpy #0
		bne lp
		ldx nuum
		ldy nuum+1
		rts
		.pend
;-----------------------------------
getnum2		.proc
		ldx #3
		lda #0
-		sta nuum,x
		dex
		bpl -
lp		lda header,y
		eor #$30
		cmp #10
		blt jo
		eor #$30^$40
		beq +
		cmp #7
		bge +
		adc #9
jo		pha
		lda #%11111000
-		ldx #252
-		rol nuum-252,x
		inx
		bne -
		lsr
		bcc --
		pla
		ora nuum
		sta nuum
+		iny
		tya
		and #7
		bne lp
		ldx nuum
		ldy nuum+1
		rts
		.pend
;-----------------------------------
topetscii	.proc
		ldx #0
		ldy #0
-		lda header.name,x
		beq +
		jsr conv
+		sta header.name,y
		ora #0
		beq +
		iny
		beq +
		inx
		bne -
+		rts

conv		cmp #32			; space-@
		blt +
		cmp #65
		bge +
		rts

+		cmp #65			;A-Z
		blt +
		cmp #91
		bge +
		ora #$80
		rts

+		cmp #97			;a-z
		blt +
		cmp #123
		bge +
		and #$df
		rts

+		cmp #192		;utf8?
		bge +
ill		lda #32
		rts

+		cmp #%11100000		;2 byte
                bcs +
                and #$1f
		sta utf+1
		jsr get
		asl a
		bcc ill
		asl a
		bcs ill
		lsr utf+1
                ror
		lsr utf+1
                ror
		sta utf
		jmp fin

+               cmp #%11110000
		bcs +
		sta utf+1
		jsr get
		sta utf
		jsr get
		asl a
		bcc ill
		asl a
		bcs ill
		asl a
		rol utf
		bcc ill
                asl
		rol utf
		bcs ill
-		asl a
		rol utf
		rol utf+1
		bcs -
		gcc fin

+               cmp #%11111000
                bcc a1
                cmp #%11111100
                bcc a2
		jsr get
a2		jsr get
a1		jsr get
		jsr get
		jsr get
		jmp ill

fin		lda utf+1
		cmp #$e1
		bne +
		lda utf
		beq ill
		rts
+
		stx xt
		ldx #size(char)-1
-		lda low,x
		cmp utf
		bne +
		lda high,x
		cmp utf+1
		bne +
		lda char,x
		ldx xt
		rts

+		dex
		bpl -
                ldx xt
		lda #32
		rts

get		inx
		lda header.name,x
		bne +
		pla
		pla
		dex
		lda #32
+		rts

		.section bss
xt		.byte ?
		.send bss

		.section data
list		= [$2215, $2f, $5b, $5b, $a3, $5c, $5d, $5d, $2191, $5e, $2190, $5f, $a0, $a0, $258c, $a1, $2584, $a2, $2594, $a3, $5f, $a4, $258f, $a5, $2592, $a6, $251c, $ab, $2597, $ac, $2514, $ad, $2510, $ae, $2582, $af, $250c, $b0, $2534, $b1, $252c, $b2, $2524, $b3, $258e, $b4, $258d, $b5, $2583, $b9, $2713, $ba, $2596, $bb, $259d, $bc, $2518, $bd, $2598, $be, $259a, $bf, $2500, $c0, $253c, $db, $2502, $dd]
low		.byte <list[::2]
high		.byte >list[::2]
char		.byte list[1::2]
		.send data
		.pend
		.send code
;-----------------------------------

		.section bss
utf		.word ?
nuum		.dword ?
currentpos	.dword ?
pad		.byte ?
pad2		.byte ?

		.union
header		.dstruct header_s
header2		.dstruct header2_s
header3		.dstruct header3_s
		.endu

name		.fill 256

		.align 256
buffer		.fill $8000
		.send bss
